import asyncio
from typing import AsyncGenerator, Generator, TypeVar

# A generic type variable for the items being yielded
T = TypeVar("T")


def async_to_sync_generator(async_gen: AsyncGenerator[T, None]) -> Generator[T, None, None]:
    """
    Converts an asynchronous generator into a synchronous generator.

    This function runs an event loop to run the async generator and yields its
    results in a synchronous manner. It properly handles cleanup when the
    sync generator is closed early.

    Args:
        async_gen: The asynchronous generator to convert.

    Yields:
        The items from the asynchronous generator.
    """
    loop = None
    async_iterator = None
    loop_was_created = False

    try:
        # Get or create an event loop
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # No running loop, create a new one
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop_was_created = True

        # Get the asynchronous iterator from the async generator
        async_iterator = async_gen.__aiter__()

        while True:
            try:
                # Run the __anext__() coroutine to get the next item.
                # loop.run_until_complete() blocks until the coroutine is done.
                item = loop.run_until_complete(async_iterator.__anext__())
                yield item
            except StopAsyncIteration:
                # The async generator is exhausted, so we break the loop.
                break
            except GeneratorExit:
                # The sync generator was closed, clean up the async generator
                print("Sync generator closed, cleaning up async generator")
                try:
                    if async_iterator is not None:
                        # Close the async generator
                        close_coro = async_iterator.aclose()  # type: ignore[attr-defined]
                        loop.run_until_complete(close_coro)
                except Exception as e:
                    print(f"Error during async generator cleanup: {e}")
                raise
    except Exception as e:
        # Handle any other exceptions during cleanup
        if async_iterator is not None:
            try:
                print(f"Exception occurred, cleaning up async generator: {e}")
                close_coro = async_iterator.aclose()  # type: ignore[attr-defined]
                if loop is not None:
                    loop.run_until_complete(close_coro)
            except Exception as cleanup_error:
                print(f"Error during exception cleanup: {cleanup_error}")
        raise
    finally:
        # Clean up the event loop if we created it
        if loop_was_created and loop is not None:
            try:
                # First, give running async generators a chance to finish their
                # ``aclose`` cleanup logic.  This prevents warnings like
                # "Task was destroyed but it is pending!" that can appear when
                # the loop is closed with unfinished async generator tasks.
                loop.run_until_complete(loop.shutdown_asyncgens())

                # Next, cancel any remaining pending tasks on the loop so that
                # we do not leave them dangling when the loop is closed.
                pending_tasks = [t for t in asyncio.all_tasks(loop) if not t.done()]
                for task in pending_tasks:
                    task.cancel()
                if pending_tasks:
                    loop.run_until_complete(asyncio.gather(*pending_tasks, return_exceptions=True))
            finally:
                try:
                    loop.close()
                except Exception as e:
                    print(f"Error closing event loop: {e}")



import json
from importlib import resources
from typing import Any, Dict

from .domain import ViteComponentsAssets


def get_vite_manifest(package_name: str) -> Dict[str, Any]:
    try:
        manifest_path = resources.files(package_name) / "static" / ".vite" / "manifest.json"

        if manifest_path.is_file():
            with manifest_path.open() as f:
                manifest_data = json.load(f)
                if isinstance(manifest_data, dict):
                    return manifest_data
    except (FileNotFoundError, AttributeError):
        pass
    return {}


def get_vite_components_assets(package_name: str) -> ViteComponentsAssets:
    manifest = get_vite_manifest(package_name)
    js_files = []
    css_files = []
    for source in manifest.values():
        if source.get("isDynamicEntry", False):
            continue
        filename = source["file"]
        if filename.endswith(".js"):
            js_files.append(filename)
        elif filename.endswith(".css"):
            css_files.append(filename)
        _css_files = source.get("css", [])
        for css_file in _css_files:
            css_files.append(css_file)
    return ViteComponentsAssets(
        package_name=package_name,
        js_files=js_files,
        css_files=css_files,
    )



import asyncio
from typing import Any, Iterator, MutableMapping, Optional

from routelit.exceptions import StopException


class PropertyDict:
    """
    A dictionary that can be accessed as attributes.
    Example:
    ```python
    session_state = PropertyDict({"name": "John"})
    print(session_state.name)  # "John"
    print(session_state["name"])  # "John"
    session_state.name = "Jane"
    print(session_state.name)  # "Jane"
    print(session_state["name"])  # "Jane"
    del session_state.name
    print(session_state.name)  # None
    print(session_state["name"])  # None
    ```
    """

    def __init__(
        self, initial_dict: Optional[MutableMapping[str, Any]] = None, cancel_event: Optional[asyncio.Event] = None
    ):
        # Initialize private attributes directly to avoid __setattr__ recursion
        super().__setattr__("_data", initial_dict if initial_dict is not None else {})
        super().__setattr__("_cancel_event", cancel_event)

    def _maybe_check_cancel(self) -> None:
        if self._cancel_event and self._cancel_event.is_set():
            raise StopException("PropertyDict cancelled")

    def __getattr__(self, name: str) -> Any:
        self._maybe_check_cancel()
        try:
            return self._data[name]
        except KeyError:
            return None

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        else:
            self.__setitem__(name, value)

    def __repr__(self) -> str:
        return f"PropertyDict({self._data!r})"

    def __str__(self) -> str:
        return str(self._data)

    def __getitem__(self, key: str) -> Any:
        self._maybe_check_cancel()
        return self._data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._maybe_check_cancel()
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        self.pop(key, None)

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def __contains__(self, key: str) -> bool:
        return key in self._data

    def pop(self, key: str, *args: Any) -> Any:
        self._maybe_check_cancel()
        return self._data.pop(key, *args)

    def get(self, key: str, default: Any = None) -> Any:
        self._maybe_check_cancel()
        return self._data.get(key, default)

    def get_data(self) -> MutableMapping[str, Any]:
        return self._data  # type: ignore[no-any-return]

    def update(self, other: MutableMapping[str, Any]) -> None:
        """Update the dictionary with key/value pairs from other."""
        self._maybe_check_cancel()
        self._data.update(other)

    def keys(self) -> Iterator[str]:
        """Return an iterator over the dictionary's keys."""
        return self._data.keys()  # type: ignore[no-any-return]

    def values(self) -> Iterator[Any]:
        """Return an iterator over the dictionary's values."""
        return self._data.values()  # type: ignore[no-any-return]

    def items(self) -> Iterator[tuple[str, Any]]:
        """Return an iterator over the dictionary's (key, value) pairs."""
        return self._data.items()  # type: ignore[no-any-return]
